{
  open MLParser

  let keyword_table =
    Hashtbl.create 53

  let keywords = [
      "let", LET;
      "in", IN;
      "fun", FUN;
      "let", LET;
      "with", WITH;
      "match", MATCH;
      "end", END;
      "type", TYPE;
      "of", OF;
      "some", SOME;
    ]

  let _ =
    List.iter
      (fun (kwd, tok) -> Hashtbl.add keyword_table kwd tok)
      keywords
}

let identchar = ['a'-'z' 'A'-'Z' '0'-'9' '_']
let lident = ['a'-'z'] identchar*
let uident = ['A'-'Z'] identchar*

let white = [' ' '\t']+
let newline = '\r' | '\n' | "\r\n"

rule read =
  parse
  | white
  | newline     { read lexbuf }
  | lident as id { try Hashtbl.find keyword_table id
                    with Not_found ->  LIDENT id }
  | uident as id { UIDENT id }
  | "'" lident as id { QIDENT id }
  | "->"        { ARROW }
  | '('		{ LPAR }
  | ')'		{ RPAR }
  | '{'		{ LBRACE }
  | '}'		{ RBRACE }
  | '*'		{ STAR }
  | ','		{ COMMA }
  | '_'		{ WILDCARD }
  | '|'		{ PIPE }
  | '='		{ EQ }
  | ":"		{ COLON }
  | '.'		{ PERIOD }
  | eof		{ EOF }
  | _ as c	{ failwith (Printf.sprintf "Unexpected character during lexing: %c" c) }