open Client

(* A few manually constructed terms. *)

let x =
  ML.Var "x"

let y =
  ML.Var "y"

let id =
  ML.Abs ("x", x)

let delta =
  ML.Abs ("x", ML.App (x, x))

let deltadelta =
  ML.App (delta, delta)

let idid =
  ML.App (id, id)

let k =
  ML.Abs ("x", ML.Abs ("y", x))

let genid =
  ML.Let ("x", id, x)

let genidid =
  ML.Let ("x", id, ML.App (x, x))

let genkidid =
  ML.Let ("x", ML.App (k, id), ML.App (x, id))

let genkidid2 =
  ML.Let ("x", ML.App (ML.App (k, id), id), x)

let app_pair = (* ill-typed *)
  ML.App (ML.Tuple [id; id], id)

let unit =
  ML.Tuple []

let abs_match_with =
  ML.(
    Abs(
      "x",
      Match(
        Tuple[],
        [ (PTuple[], Tuple[]) ]
      )
    )
  )

(* option *)
let option_env =

  (* type 'a option = None | Some of 'a *)
  let option_typedecl =
    let open Datatype in {
    name = Type "option";
    type_params = [ "a" ];
    data_kind = Variant;
    labels_descr = [
      {
        label_name = Label"None";
        type_name = Type "option";
        arg_type = ML.TyProduct [];
      } ; {
        label_name = Label "Some";
        type_name = Type "option";
        arg_type = ML.TyVar "a";
      }
    ]
  } in
  Datatype.Env.add_decl Datatype.Env.empty option_typedecl

let none = ML.Variant ( Datatype.Label "None" , ML.Tuple [] )

let some =
  ML.Variant (
    Datatype.Label "Some",
    id
  )

let match_none = ML.(
    Match (none, [
      PVariant (Datatype.Label "None", PTuple []), none ;
      PVariant (Datatype.Label "Some", PVar "x"), x ;
    ])
  )

let match_some = ML.(
    Match (some, [
      PVariant (Datatype.Label "None", PTuple []), none ;
      PVariant (Datatype.Label "Some", PWildcard), none
    ])
  )

let match_some_annotated = ML.(
    Match (some, [
      ( PVariant (Label "None", PTuple []), none );
      ( PAnnot (PVariant (Label "Some", PWildcard),
               (["'a"], TyConstr (Type "option", [ TyVar "'a" ]))),
        none );
    ])
  )

(* list *)
let list_env =

  (* type 'a list = Nil | Cons of 'a * 'a list *)
  let list_typedecl =
    let open Datatype in {
    name = Type "list";
    type_params = [ "a" ];
    data_kind = Variant;
    labels_descr = [
      {
        label_name = Label "Nil";
        type_name = Type "list";
        arg_type = ML.TyProduct [];
      } ; {
        label_name = Label "Cons";
        type_name = Type "list";
        arg_type = ML.(TyProduct [ TyVar "a" ; TyConstr (Type "list", [ TyVar "a" ]) ]);
      }
    ]
  } in

  Datatype.Env.add_decl option_env list_typedecl

let nil = ML.Variant ( Datatype.Label "Nil" , ML.Tuple [] )

let list =
  ML.Variant (
    Datatype.Label "Cons",
    ML.Tuple [ id ; nil ]
  )

let list_annotated =
  let open ML in
  Annot (
    Variant (
      Label "Cons",
      Tuple [ Annot (id, (["a"], TyArrow (TyVar "a", TyVar "a")));
              nil ]
    ),
    (["a"; "b"], TyConstr (Type "list", [TyArrow (TyVar "a", TyVar "b")]))
  )

(* tree *)
let tree_env =

  (* type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree *)
  let tree_typedecl =
    let open Datatype in {
    name = Type "tree";
    type_params = [ "a" ];
    data_kind = Variant;
    labels_descr = [
      {
        label_name = Label "Leaf";
        type_name = Type "tree";
        arg_type = ML.TyProduct []
      } ; {
        label_name = Label "Node";
        type_name = Type "tree";
        arg_type = ML.(TyProduct [
          TyConstr (Type "tree", [ TyVar "a" ]);
          TyVar "a";
          TyConstr (Type "tree", [ TyVar "a" ]);
        ])
      }
    ];
  } in

  Datatype.Env.add_decl list_env tree_typedecl

let leaf = ML.Variant ( Datatype.Label "Leaf" , ML.Tuple [] )

let node =
  ML.Variant (
    Datatype.Label "Node",
    ML.Tuple [
      leaf ;
      id ;
      leaf ;
    ]
  )

let () =
  assert Test.(Log.with_log CheckML.test Datatype.Env.empty idid);
  assert Test.(Log.with_log CheckML.test Datatype.Env.empty genid);
  assert Test.(Log.with_log CheckML.test Datatype.Env.empty genidid);
  assert Test.(Log.with_log CheckML.test Datatype.Env.empty genkidid);
  assert Test.(Log.with_log CheckML.test Datatype.Env.empty genkidid2);
  assert Test.(Log.with_log CheckML.test option_env none);
  assert Test.(Log.with_log CheckML.test option_env some);
  assert Test.(Log.with_log CheckML.test list_env nil);
  assert Test.(Log.with_log CheckML.test list_env list);
  assert Test.(Log.with_log CheckML.test list_env list_annotated);
  assert Test.(Log.with_log CheckML.test tree_env leaf);
  assert Test.(Log.with_log CheckML.test tree_env node);
  assert Test.(Log.with_log CheckML.test Datatype.Env.empty abs_match_with);
  assert Test.(Log.with_log CheckML.test option_env match_none);
  assert Test.(Log.with_log CheckML.test option_env match_some);
  assert Test.(Log.with_log CheckML.test option_env match_some_annotated);
  (* we include some printing below because Dune
     appears to only show tests that show some output. *)
  print_endline "TestML: all tests passed."


(*************************************************************************)

(* Main parsing function *)
let parse = MLParser.prog MLLexer.read

(* Currently unused *)
let ast_from_string s =
  let lexbuf = Lexing.from_string s in
  parse lexbuf

exception ParsingError

let letify (typedecl_env, xts) =
  let rec aux xts =
    match xts with
    | [] ->
        assert false
    | [(_, t)] ->
        t
    | (x, t) :: xts ->
        ML.Let (x, t, aux xts)
  in
  (typedecl_env, aux xts)

let test_from_string typedecl s expected on_same on_different on_error =
  let s = typedecl ^ "\nlet _ = " ^ s in
  let t =
    try Ok (ast_from_string s |> letify)
    with e -> print_endline s; raise e (*Error ()*)
  in
  match t with
  | Ok (datatype_env, t) ->
     if t = expected then
       begin
         assert Test.(Log.with_log CheckML.test datatype_env t);
         on_same ();
       end
     else
       on_different ()
  | Error () ->
     on_error ()

let test_ok ?(typedecl="") s expected =
  test_from_string typedecl s expected
    (fun () -> ())
    (fun () -> prerr_endline ("different: " ^ s); raise ParsingError)
    (fun () -> prerr_endline ("error: " ^ s); raise ParsingError)

let test_error _s _expected = (*
  test_from_string "" s expected
    (fun () -> prerr_endline s; raise ParsingError)
    (fun () -> ())
    (fun () -> ()) *) ()

let test_id =
  test_ok "fun x -> x" id

let test_delta_delta_error =
  test_error "(fun x -> x x (fun x -> x x)" deltadelta

let test_idid =
  test_ok "(fun x -> x) (fun x -> x)" idid

let test_idid_error =
  test_error "fun x -> x fun x -> x" idid

let test_unit =
  test_ok "()" unit

let test_abs_match_with =
  test_ok "fun x -> match () with () -> () end" abs_match_with

let test_let =
  test_ok "let y = fun x -> x in ()" (ML.Let("y", id, unit))

let test_let_prod_singleton =
  test_ok "let (y,) = (fun x -> x,) in ()"
    (ML.LetProd (["y"], ML.Tuple [id], unit))

let test_let_prod =
  test_ok "let (y,z) = (fun x -> x, ()) in ()"
  (ML.LetProd (["y";"z"], ML.Tuple [id;unit], unit))

let test_singleton =
  test_ok
    "(fun x -> x,)"
    (ML.Tuple [id])

let test_pair_tuple =
  test_ok
    "(fun x -> x, fun x -> x)"
    (ML.Tuple [id; id])

let test_none =
  test_ok
    ~typedecl:"type option 'a = None of {} | Some of 'a"
    "None"
    none

let test_none_unit =
  test_ok
    ~typedecl:"type option 'a = None of {} | Some of 'a"
    "None ()"
    none

let test_some =
  test_ok
    ~typedecl:"type option 'a = None of {} | Some of 'a"
    "Some (fun x -> x)"
    some

let test_some_pair =
  test_ok
    ~typedecl:"type option 'a = None of {} | Some of 'a"
    "Some (fun x -> x, fun x -> x)"
    (ML.Variant (Datatype.Label "Some", ML.Tuple [id;id]))

let test_list_nil =
  test_ok
    ~typedecl:"type list 'a = Nil of {} | Cons of {'a * list 'a}"
    "Nil"
    nil

let test_list_cons =
  test_ok
    ~typedecl:"type list 'a = Nil of {} | Cons of {'a * list 'a}"
    "Cons (fun x -> x, Nil)"
    list

let test_arrow =
  test_ok
    ~typedecl:"type func 'a 'b = Func of 'a -> 'b"
    "Func (fun x -> x)"
    (ML.Variant (Datatype.Label "Func", id))

let test_match_tuple =
  test_ok
    "match (fun x -> x, ()) with (f, ()) -> f end"
    (ML.Match
       (ML.Tuple [id;unit],
        [ (PTuple[PVar "f"; PTuple[]], Var "f") ]
    ))

let test_match_none =
  test_ok
    ~typedecl:"type option 'a = None of {} | Some of 'a"
    "match None () with
     | None () -> None ()
     | Some x -> x
     end"
  match_none

let test_match_some =
  test_ok
    ~typedecl:"type option 'a = None of {} | Some of 'a"
    "match Some (fun x -> x) with
     | None () -> None ()
     | Some _ -> None ()
     end"
  match_some

let test_match_some_annotated =
  test_ok
    ~typedecl:"type option 'a = None of {} | Some of 'a"
    "match Some (fun x -> x) with
     | None () -> None ()
     | (Some _ : some 'a. option 'a) -> None ()
     end"
  match_some_annotated
